# -*- coding: utf-8 -*-
"""
Normalizing of HMHospitales COVID19 database 01
This version contains the same method, with some new features:
    - Use of Cross Validation
    - 

@author: varsha
"""
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from sklearn.preprocessing import LabelEncoder
from sklearn.preprocessing import MinMaxScaler
from sklearn.ensemble import RandomForestClassifier
from sklearn import metrics
from sklearn.metrics import confusion_matrix, plot_confusion_matrix, recall_score
from sklearn.model_selection import train_test_split, KFold, cross_validate
from keras.metrics import MeanIoU
import numpy as np
from datetime import datetime

df = pd.read_csv('../Database/HM_Hospitales/csv/01.csv')

print(df.describe().T)

sns.displot(df['DIAG ING/INPAT'], kde=False)

discharge_plot = sns.countplot(x="MOTIVO_ALTA/DESTINY_DISCHARGE_ING", data=df)
discharge_plot.set_xticklabels(discharge_plot.get_xticklabels(), rotation=40, ha="right")
plt.tight_layout()
plt.show()

#Print correlation matrix
print(df.corr())
corrMatrix = df.corr()
fig, ax = plt.subplots(figsize=(10,10))         # Sample figsize in inches
sns.heatmap(corrMatrix, annot=False, linewidths=.5, ax=ax)

print(df.isnull().sum())

#Drop columns with too many null values (~2000 out of 2300)
df_clean = df.drop(["F_ENTRADA_UC/ICU_DATE_IN","F_SALIDA_UCI/ICU_DATE_OUT"], axis=1)

#'ICU_DAYS' has 2106 null values, meaning 0 days in UCI, so these values will be filled with '0'
df_clean['UCI_DIAS/ICU_DAYS'].fillna(0, inplace=True)
print(df_clean.isnull().sum())

#Drop (not needed) columns with some null values
df_clean = df_clean.drop(["HORA/TIME_CONSTANT_PRIMERA/FIRST_URG/EMERG","HORA/TIME_CONSTANT_ULTIMA/LAST_URG/EMERG",
                          "F_INGRESO/ADMISSION_DATE_URG/EMERG","HORA/TIME_ADMISION/ADMISSION_URG/EMERG", 
                          "F_ALTA/DISCHARGE_DATE_ING", "F_INGRESO/ADMISSION_D_ING/INPAT"], 
                         axis=1)


df_clean['MOTIVO_ALTA/DESTINY_DISCHARGE_ING'].fillna("NO ALTA", inplace=True)
df_clean['MOTIVO_ALTA/DESTINY_DISCHARGE_ING'].value_counts()

#Drop all rows with at least one null value
df_clean = df_clean.dropna()
print(df_clean.isnull().sum())

print(df_clean['MOTIVO_ALTA/DESTINY_DISCHARGE_ING'].value_counts())
## Drop rows where Discharge reason is "volunteer" or a transfer
df_clean = df_clean.drop(df_clean[
    (df_clean["MOTIVO_ALTA/DESTINY_DISCHARGE_ING"]== "Alta Voluntaria") |
    (df_clean["MOTIVO_ALTA/DESTINY_DISCHARGE_ING"]== "Traslado al Hospital") |
    (df_clean["MOTIVO_ALTA/DESTINY_DISCHARGE_ING"] == "Traslado a un Centro Sociosanitario")].index)


#Change column names to be readable
df_clean = df_clean.rename(columns={'PATIENT ID':'id',
                         'EDAD/AGE':'age',
                         'SEXO/SEX':'sex',
                         'DIAG ING/INPAT':'covid_diagnosis',
                         'UCI_DIAS/ICU_DAYS':'icu_days',
                         'MOTIVO_ALTA/DESTINY_DISCHARGE_ING':'discharge_reason',
                         'ESPECIALIDAD/DEPARTMENT_URG/EMERG':'department',
                         'DIAG_URG/EMERG':'emergency_diagnosis',
                         'DESTINO/DESTINY_URG/EMERG':'emergency_dept',
                         'TEMP_PRIMERA/FIRST_URG/EMERG':'em_first_temperature',
                         'FC/HR_PRIMERA/FIRST_URG/EMERG':'em_first_hr',
                         'GLU_PRIMERA/FIRST_URG/EMERG':'em_first_glucemia',
                         'SAT_02_PRIMERA/FIRST_URG/EMERG':'em_first_OxSaturation',
                         'TA_MAX_PRIMERA/FIRST/EMERG_URG':'em_first_max_blood_pressure',
                         'TA_MIN_PRIMERA/FIRST_URG/EMERG':'em_first_min_blood_pressure',
                         'FC/HR_ULTIMA/LAST_URG/EMERG':'em_last_hr',
                         'TEMP_ULTIMA/LAST_URG/EMERG':'em_last_temperature',
                         'GLU_ULTIMA/LAST_URG/EMERG':'em_last_glucemia',
                         'SAT_02_ULTIMA/LAST_URG/EMERG':'em_last_OxSaturation',
                         'TA_MAX_ULTIMA/LAST_URGEMERG':'em_last_max_blood_pressure',
                         'TA_MIN_ULTIMA/LAST_URG/EMERG':'em_last_min_blood_pressure'})
print(df_clean.dtypes)
print(df_clean['discharge_reason'].value_counts())

df_two_binary = df_clean

#Change text (object) variables to be numbers
#Randomly done by LabelEncoder (we do this to because text data is more complicated to analyse)

df_clean['sex'] = LabelEncoder().fit_transform(df_clean['sex'])
df_clean['sex'].value_counts()

df_clean['covid_diagnosis'] = LabelEncoder().fit_transform(df_clean['covid_diagnosis'])
df_clean['covid_diagnosis'].value_counts()

df_clean['department'] = LabelEncoder().fit_transform(df_clean['department'])
df_clean['department'].value_counts()

df_clean['emergency_diagnosis']= LabelEncoder().fit_transform(df_clean['emergency_diagnosis'])
df_clean['emergency_diagnosis'].value_counts()

df_clean['emergency_dept'] = LabelEncoder().fit_transform(df_clean['emergency_dept'])
df_clean['emergency_dept'].value_counts()

df_clean['discharge_reason'] = LabelEncoder().fit_transform(df_clean['discharge_reason'])
df_clean['discharge_reason'].value_counts()

print(df_clean.dtypes)
print(df_clean.describe().T)

# DEFINING X AND Y

# Y = variables to be predicted (patient progression)
Y = df_clean["discharge_reason"].values
# X = information we have about the patient
X = df_clean.drop(labels = ["id", "discharge_reason"], axis=1)
features_list = list(X.columns)

#Scale data to be between 0 and 1
scaler = MinMaxScaler()
scaler.fit(X)
X = scaler.transform(X)


########### INITIAL PREDICTION ###########
#Split dataframe data to get train and test variables (80% train and 20% test)
X_train, X_test, y_train, y_test = train_test_split(X, Y, test_size=0.2, random_state=32)

#Use RANDOM FOREST CLASSIFIER to make predictions
model = RandomForestClassifier(random_state = 32)
# Train the model on training data
model.fit(X_train, y_train)
prediction = model.predict(X_test)


#Prediction is done, now let's check the model's accuracy
print("Accuracy = ", metrics.accuracy_score(y_test, prediction))

#Confusion matrix
cm= confusion_matrix(y_test, prediction)
print(cm)

plot_confusion_matrix(model, X_test, y_test)
plt.show()

#Print individual accuracy values for each class, based on the confusion matrix
df['MOTIVO_ALTA/DESTINY_DISCHARGE_ING'].value_counts()
print("Domicilio = ",   cm[0,0] / (cm[0,0]+cm[1,0]+cm[2,0]))
print("Fallecimiento = ",   cm[1,1] / (cm[0,1]+cm[1,1]+cm[2,1]))
print("NO ALTA = ",   cm[2,2] / (cm[0,2]+cm[1,2]+cm[2,2]))

#Feature importances
feature_imp = pd.Series(model.feature_importances_, index=features_list).sort_values(ascending=False)
print(feature_imp)

#Calculate accuracy using Intersection over Union
num_classes = 3
IOU_RF = MeanIoU(num_classes=num_classes)  
IOU_RF.update_state(y_test, prediction)
print("Mean IoU for Random Forest = ", IOU_RF.result().numpy())

values = np.array(IOU_RF.get_weights()).reshape(num_classes, num_classes)
print(values)
class1_IoU = values[0,0]/(values[0,0] + values[0,1] + values[0,2] + values[1,0] + values[2,0])
class2_IoU = values[1,1]/(values[1,1] + values[1,0] + values[1,2] + values[0,1] + values[2,1])
class3_IoU = values[2,2]/(values[2,2] + values[2,0] + values[2,1] + values[0,2] + values[1,2])

print("IoU for class 1 [Domicilio] is: ", class1_IoU)
print("IoU for class 2 [Fallecimiento] is: ", class2_IoU)
print("IoU for class 3 [NO ALTA] is: ", class3_IoU)

# RANDOM FOREST ACCURACY = 0.76 (vs) 0.38 = IoU
# Lets balance the data, BUT KEEPING THE SAME test DATA

from sklearn.utils import resample

train_set = np.concatenate((X_train, y_train[:,None]), axis=1)

dataset1 = train_set[train_set[:,19] == 0] #Domicilio
dataset2 = train_set[train_set[:,19] == 1] #Fallecimiento
dataset3 = train_set[train_set[:,19] == 2] #NO ALTA (pendiente = testing)
df_clean["discharge_reason"].value_counts()
#Let's choose a sample value of 1000

df1 = resample(dataset1, 
                                 replace=True,     # sample with replacement
                                 n_samples=1000,    # to match average class
                                 random_state=32)

df2 = resample(dataset2, 
                                 replace=True,     # sample with replacement
                                 n_samples=1000,    # to match average class
                                 random_state=32)

df3 = resample(dataset3, 
                                 replace=True,     # sample with replacement
                                 n_samples=1000,    # to match average class
                                 random_state=32)


dataset_resampled = np.concatenate([df1,df2,df3])
#Now that only train data is copied, we need to reassign them again to 'X' and 'Y'
X_train_resampled = np.delete(dataset_resampled, [19], axis=1)
y_train_resampled = dataset_resampled[:,19]

#Now the same model is used but without the testing data being already used for training

RF_model_resampled = RandomForestClassifier(random_state = 32)
 
## Train the model on training data
start = datetime.now() 

RF_model_resampled.fit(X_train_resampled, y_train_resampled)
stop = datetime.now()


#Execution time of the model 
execution_time_RF_resampled = stop-start 
print("Random Forest resampled execution time is: ", execution_time_RF_resampled)

prediction_RF_resampled = RF_model_resampled.predict(X_test)
print ("Accuracy of Random Forest resampled = ", metrics.accuracy_score(y_test, prediction_RF_resampled))

# RF accuracy is 0.7 but what about IoU?
num_classes = 3
IOU_RF = MeanIoU(num_classes=num_classes)  
IOU_RF.update_state(y_test, prediction_RF_resampled)
print("Mean IoU for Random Forest = ", IOU_RF.result().numpy())

values = np.array(IOU_RF.get_weights()).reshape(num_classes, num_classes)
print(values)

# IoU accuracy is now 0.37, it improved VERY LITTLE, let's see for specific classes
class1_IoU = values[0,0]/(values[0,0] + values[0,1] + values[0,2] + values[1,0] + values[2,0])
class2_IoU = values[1,1]/(values[1,1] + values[1,0] + values[1,2] + values[0,1] + values[2,1])
class3_IoU = values[2,2]/(values[2,2] + values[2,0] + values[2,1] + values[0,2] + values[1,2])

print("IoU for class 1 is: ", class1_IoU)
print("IoU for class 2 is: ", class2_IoU)
print("IoU for class 3 is: ", class3_IoU)

## CROSS VALIDATION ##
from sklearn.model_selection import cross_val_score
# Let's start using the number of folds from '5'
RF_model_folds = RandomForestClassifier(n_estimators = 25, random_state = 32)
scores = cross_val_score(RF_model_folds, X_train, y_train, cv=10)
print(scores)
print("%0.2f accuracy with a standard deviation of %0.2f" % (scores.mean(), scores.std()))
#### RESULT: 75% (without resampling any data, best results so far, with: cv = 5, 10, 30 and 100)

############################## SMOTE v1 ################################

from imblearn.over_sampling import SMOTE
oversampler = SMOTE(random_state=32)
X_over, y_over = oversampler.fit_resample(X_train, y_train)

oversampled_plot = sns.countplot(y_over)
plt.tight_layout()
plt.show()

RF_model_over = RandomForestClassifier(random_state = 32)

## OVERSAMPLING WITHOUT FOLDS ##
RF_model_over.fit(X_over, y_over)

prediction_RF_over = RF_model_over.predict(X_test)
print ("Accuracy of Random Forest resampled = ", metrics.accuracy_score(y_test, prediction_RF_over))
#It went from 0.7 to 0.71. Let's see for each class

num_classes = 3
IOU_RF = MeanIoU(num_classes=num_classes)  
IOU_RF.update_state(y_test, prediction_RF_over)
print("Mean IoU for Random Forest = ", IOU_RF.result().numpy())

values = np.array(IOU_RF.get_weights()).reshape(num_classes, num_classes)
print(values)

# IoU accuracy is now 0.37, it is the same as the resampled version, let's see for specific classes
class1_IoU = values[0,0]/(values[0,0] + values[0,1] + values[0,2] + values[1,0] + values[2,0])
class2_IoU = values[1,1]/(values[1,1] + values[1,0] + values[1,2] + values[0,1] + values[2,1])
class3_IoU = values[2,2]/(values[2,2] + values[2,0] + values[2,1] + values[0,2] + values[1,2])

print("IoU for class 1 is: ", class1_IoU)
print("IoU for class 2 is: ", class2_IoU)
print("IoU for class 3 is: ", class3_IoU)

##############################################################################################

## CROSS-VAL WITH OVERSAMPLED DATA ##
from sklearn.model_selection import cross_val_score

manual_cm = np.array([[0,0,0],[0,0,0],[0,0,0]])

def confusion_matrix_scorer(model, X, y):
    global manual_cm
    y_pred = model.predict(X)
    fold_cm = confusion_matrix(y, y_pred)
    score = (((fold_cm[0,0]/(fold_cm[0,0] + fold_cm[1,0] + fold_cm[2,0])) +
             (fold_cm[1,1]/(fold_cm[1,1] + fold_cm[0,1] + fold_cm[2,1]))+
             (fold_cm[2,2]/(fold_cm[2,2] + fold_cm[0,2] + fold_cm[1,2])))/ 3)
    manual_cm = np.add(manual_cm, fold_cm)
    return score

RF_model_folds_over = RandomForestClassifier(random_state = 32)

scores = cross_validate(RF_model_folds_over, X_over, y_over, cv=10, scoring=confusion_matrix_scorer)
print("Mean is: ", np.mean(scores['test_score']))
# IoU mean
class1_IoU = manual_cm[0,0]/(manual_cm[0,0] + manual_cm[0,1] + manual_cm[0,2] + manual_cm[1,0] + manual_cm[2,0])
class2_IoU = manual_cm[1,1]/(manual_cm[1,1] + values[1,0] + manual_cm[1,2] + manual_cm[0,1] + manual_cm[2,1])
class3_IoU = manual_cm[2,2]/(manual_cm[2,2] + values[2,0] + manual_cm[2,1] + manual_cm[0,2] + manual_cm[1,2])
print("IoU mean is: ", (class1_IoU + class2_IoU + class3_IoU) / 3)
print("[ ", class1_IoU, ", ", class2_IoU, ", ", class3_IoU, " ]")
#print("%0.2f accuracy with a standard deviation of %0.2f" % (scores.mean(), scores.std()))
 
# Adding 'test' to cross-val + oversampled data
X_full = np.append(X_over, X_test, axis=0)
y_full = np.append(y_over, y_test)
scores = cross_validate(RF_model_folds_over, X_full, y_full, cv=10, scoring=confusion_matrix_scorer)
print("Mean is: ", np.mean(scores['test_score']))
# IoU mean
class1_IoU = manual_cm[0,0]/(manual_cm[0,0] + manual_cm[0,1] + manual_cm[0,2] + manual_cm[1,0] + manual_cm[2,0])
class2_IoU = manual_cm[1,1]/(manual_cm[1,1] + values[1,0] + manual_cm[1,2] + manual_cm[0,1] + manual_cm[2,1])
class3_IoU = manual_cm[2,2]/(manual_cm[2,2] + values[2,0] + manual_cm[2,1] + manual_cm[0,2] + manual_cm[1,2])
print("IoU mean is: ", (class1_IoU + class2_IoU + class3_IoU) / 3)
print("[ ", class1_IoU, ", ", class2_IoU, ", ", class3_IoU, " ]")
#print("%0.2f accuracy with a standard deviation of %0.2f" % (scores.mean(), scores.std()))


##############################################################################################

from imblearn.over_sampling import SMOTE
# SMOTE (Oversampling) for each fold
# Define model
smoter = SMOTE(random_state=32)
#smoter = RandomOverSampler(sampling_strategy='minority')
scores = []
cv = KFold(n_splits=3, random_state=32, shuffle=True)
X_train_folds = {}
y_train_folds = {}
X_val_folds = {}
y_val_folds = {}
i = 0
smote_cm = np.array([[0,0,0],[0,0,0],[0,0,0]])

for train_fold_index, val_fold_index in cv.split(X_train, y_train):
    # Get the training data
    X_train_fold, y_train_fold = X_train[train_fold_index], y_train[train_fold_index]
    # Get the validation data
    X_val_fold, y_val_fold = X_train[val_fold_index], y_train[val_fold_index]

    # Upsample only the data in the training section
    X_train_fold_upsample, y_train_fold_upsample = smoter.fit_resample(X_train_fold,
                                                                       y_train_fold)
    X_train_folds["x" + str(i)] = X_train_fold_upsample
    y_train_folds["y" + str(i)] = y_train_fold_upsample
    X_val_folds["x" + str(i)] = X_val_fold
    y_val_folds["y" + str(i)] = y_val_fold

    i += 1

def score_model(params):
    global smote_cm
    RF_model_over_cross = RandomForestClassifier(**params)
    for j in range(len(X_train_folds)):
            # Fit the model on the upsampled training data
            model_obj = RF_model_over_cross.fit(X_train_folds["x"+ str(j)], y_train_folds["y" + str(j)])
            # Score the model on the (non-upsampled) validation data
            prediction =  model_obj.predict(X_val_folds["x"+ str(j)])
            score = recall_score(y_val_folds["y" + str(j)], prediction , average=None)
            scores.append(score)
            smote_cm = np.add(smote_cm, confusion_matrix(y_val_folds["y" + str(j)], prediction))
            return scores
        
#Let's make a loop to find the best parameters for the model
score_tracker = []
for n_estimators in [25, 50, 100, 200]:
    example_params = {
        'n_estimators': n_estimators,
        'random_state': 32
    }
    example_params['recall'] = np.mean(score_model(example_params))
    score_tracker.append(example_params)
        
        
# Class acuracy of the cross-val
# Normal
class1 = smote_cm[0,0] / (smote_cm[0,0] + smote_cm[1,0] + smote_cm[2,0])
class2 = smote_cm[1,1] / (smote_cm[1,1] + smote_cm[0,1] + smote_cm[2,1]) 
class3 = smote_cm[2,2] / (smote_cm[2,2] + smote_cm[0,2] + smote_cm[1,2])
print("[ ", class1, ", ", class2, ", ", class3, " ]")
print("MEAN CROSS-VAL SMOTE: ", (class1 + class2 + class3)/3)
# IoU
class1_iou = smote_cm[0,0] / (smote_cm[0,0] + smote_cm[1,0] + smote_cm[2,0] + smote_cm[0,1] + smote_cm[0,2])
class2_iou = smote_cm[1,1] / (smote_cm[1,1] + smote_cm[0,1] + smote_cm[2,1] + smote_cm[1,0] + smote_cm[1,2]) 
class3_iou = smote_cm[2,2] / (smote_cm[2,2] + smote_cm[0,2] + smote_cm[1,2] + smote_cm[2,0] + smote_cm[2,1])
print("[ ", class1_iou, ", ", class2_iou, ", ", class3_iou, " ]")
print("MEAN CROSS-VAL SMOTE IoU:", (class1_iou + class2_iou + class3_iou)/3)

# What's the best model?
sorted(score_tracker, key=lambda x: x['recall'], reverse=True)[0]


##########################################################################################
##################################### BEST PARAMS ########################################
##########################################################################################
# Testing with best params
RF_best_params = RandomForestClassifier(n_estimators = 50, random_state=32)
RF_best_params.fit(X_over, y_over)
prediction_oversample = RF_best_params.predict(X_test)
recall_score(y_test, prediction_oversample, average='weighted')
recall_score(y_test, prediction_oversample, average=None)

# Let's check IoU
from keras.metrics import MeanIoU
num_classes = 3
IOU_RF = MeanIoU(num_classes=num_classes)  
IOU_RF.update_state(y_test, prediction_oversample)
print("Mean IoU for Random Forest = ", IOU_RF.result().numpy())

values = np.array(IOU_RF.get_weights()).reshape(num_classes, num_classes)
print(values)

class1_IoU = values[0,0]/(values[0,0] + values[0,1] + values[0,2] + values[1,0] + values[2,0])
class2_IoU = values[1,1]/(values[1,1] + values[1,0] + values[1,2] + values[0,1] + values[2,1])
class3_IoU = values[2,2]/(values[2,2] + values[2,0] + values[2,1] + values[0,2] + values[1,2])

print("IoU for class [Domicilio] is: ", class1_IoU)
print("IoU for class [Fallecimiento] is: ", class2_IoU)
print("IoU for class [No Alta] is: ", class3_IoU)


#Best params oversampling each fold
X_upsampled_folds = np.concatenate((X_train_folds["x0"], X_train_folds["x1"], X_train_folds["x2"]))
y_upsampled_folds = np.concatenate((y_train_folds["y0"], y_train_folds["y1"], y_train_folds["y2"]))

RF_best_params = RandomForestClassifier(n_estimators=50, random_state=42)
RF_best_params.fit(X_upsampled_folds, y_upsampled_folds)
prediction_fold_over = RF_best_params.predict(X_test)
recall_score(y_test, prediction_fold_over, average='weighted')
recall_score(y_test, prediction_fold_over, average=None)

# Let's check IoU
from keras.metrics import MeanIoU
num_classes = 3
IOU_RF = MeanIoU(num_classes=num_classes)  
IOU_RF.update_state(y_test, prediction_fold_over)
print("Mean IoU for Random Forest = ", IOU_RF.result().numpy())

values = np.array(IOU_RF.get_weights()).reshape(num_classes, num_classes)
print(values)

class1_IoU = values[0,0]/(values[0,0] + values[0,1] + values[0,2] + values[1,0] + values[2,0])
class2_IoU = values[1,1]/(values[1,1] + values[1,0] + values[1,2] + values[0,1] + values[2,1])
class3_IoU = values[2,2]/(values[2,2] + values[2,0] + values[2,1] + values[0,2] + values[1,2])

print("IoU for class [Domicilio] is: ", class1_IoU)
print("IoU for class [Fallecimiento] is: ", class2_IoU)
print("IoU for class [No Alta] is: ", class3_IoU)

##############################################################
        ## METHOD 2: SUPPORT VECTOR CLASSIFICATOR ##
##############################################################

# MULTI-CLASS SVM (Multi-class, linear) ON OVERSAMPLED DATA 
from sklearn import svm

SVM_oversampled = svm.LinearSVC()
SVM_oversampled.fit(X_over, y_over)

#Predict
svm_prediction = SVM_oversampled.predict(X_test)

#Accuracy
print("Accuracy SVM = ", metrics.accuracy_score(y_test, svm_prediction))

# Let's check IoU
from keras.metrics import MeanIoU
num_classes = 3
IOU_RF = MeanIoU(num_classes=num_classes)  
IOU_RF.update_state(y_test, svm_prediction)
print("Mean IoU for SVM = ", IOU_RF.result().numpy())

values = np.array(IOU_RF.get_weights()).reshape(num_classes, num_classes)
print(values)

class1_IoU = values[0,0]/(values[0,0] + values[0,1] + values[0,2] + values[1,0] + values[2,0])
class2_IoU = values[1,1]/(values[1,1] + values[1,0] + values[1,2] + values[0,1] + values[2,1])
class3_IoU = values[2,2]/(values[2,2] + values[2,0] + values[2,1] + values[0,2] + values[1,2])

print("IoU for class [Domicilio] is: ", class1_IoU)
print("IoU for class [Fallecimiento] is: ", class2_IoU)
print("IoU for class [No Alta] is: ", class3_IoU)

##############################################################
                   ## METHOD 3: KNN ##
##############################################################

from sklearn.neighbors import KNeighborsClassifier

knn_oversampled = KNeighborsClassifier()

knn_oversampled.fit(X_over, y_over)

#Predict
knn_prediction = knn_oversampled.predict(X_test)

#Accuracy
print("Accuracy KNN = ", metrics.accuracy_score(y_test, knn_prediction))

# Let's check IoU
from keras.metrics import MeanIoU
num_classes = 3
IOU_RF = MeanIoU(num_classes=num_classes)  
IOU_RF.update_state(y_test, knn_prediction)
print("Mean IoU for KNN = ", IOU_RF.result().numpy())

values = np.array(IOU_RF.get_weights()).reshape(num_classes, num_classes)
print(values)

class1_IoU = values[0,0]/(values[0,0] + values[0,1] + values[0,2] + values[1,0] + values[2,0])
class2_IoU = values[1,1]/(values[1,1] + values[1,0] + values[1,2] + values[0,1] + values[2,1])
class3_IoU = values[2,2]/(values[2,2] + values[2,0] + values[2,1] + values[0,2] + values[1,2])

print("IoU for class [Domicilio] is: ", class1_IoU)
print("IoU for class [Fallecimiento] is: ", class2_IoU)
print("IoU for class [No Alta] is: ", class3_IoU)


# BEST RESULTS: ?



############################ UNDERSAMPLING WHOLE #############################

from imblearn.under_sampling import RandomUnderSampler
rus = RandomUnderSampler(random_state=0)
X_under, y_under = rus.fit_resample(X_train, y_train)
RF_under = RandomForestClassifier(random_state=32)

RF_under.fit(X_under, y_under)
RF_under_prediction = RF_under.predict(X_test)

#Accuracy
print("Accuracy RF under = ", metrics.accuracy_score(y_test, RF_under_prediction))
#0.59 (not good)

# Let's check IoU
from keras.metrics import MeanIoU
num_classes = 3
IOU_RF = MeanIoU(num_classes=num_classes)  
IOU_RF.update_state(y_test, RF_under_prediction)
print("Mean IoU for KNN = ", IOU_RF.result().numpy())
#0.35

values = np.array(IOU_RF.get_weights()).reshape(num_classes, num_classes)
print(values)

class1_IoU = values[0,0]/(values[0,0] + values[0,1] + values[0,2] + values[1,0] + values[2,0])
class2_IoU = values[1,1]/(values[1,1] + values[1,0] + values[1,2] + values[0,1] + values[2,1])
class3_IoU = values[2,2]/(values[2,2] + values[2,0] + values[2,1] + values[0,2] + values[1,2])

print("IoU for class [Domicilio] is: ", class1_IoU)
print("IoU for class [Fallecimiento] is: ", class2_IoU)
print("IoU for class [No Alta] is: ", class3_IoU)

########################## TOP 10 IMPORTANT FEATURES ##########################

df_top10 = df_clean.drop(labels = ["em_first_min_blood_pressure", 
                                   "emergency_diagnosis",
                                   "icu_days","sex",
                                   "department",
                                   "covid_diagnosis",
                                   "em_last_glucemia",
                                   "em_first_glucemia",
                                   "emergency_dept"], axis=1)

# DEFINING X AND Y

Y_top10 = df_top10["discharge_reason"].values
X_top10 = df_top10.drop(labels = ["id", "discharge_reason"], axis=1)
features_list = list(X_top10.columns)

#Scale data to be between 0 and 1
scaler = MinMaxScaler()
scaler.fit(X_top10)
X_top10 = scaler.transform(X_top10)


########### INITIAL PREDICTION ###########
#Split dataframe data to get train and test variables (80% train and 20% test)
X_train_top10, X_test_top10, y_train_top10, y_test_top10 = train_test_split(X_top10, Y_top10, test_size=0.2, random_state=32)

#Use RANDOM FOREST CLASSIFIER to make predictions
RF_top10 = RandomForestClassifier(random_state = 32)
# Train the model on training data
RF_top10.fit(X_train_top10, y_train_top10)
prediction_top10 = RF_top10.predict(X_test_top10)


#Prediction is done, now let's check the model's accuracy
print("Accuracy = ", metrics.accuracy_score(y_test_top10, prediction_top10))

#Confusion matrix
cm= confusion_matrix(y_test_top10, prediction_top10)
print(cm)


############################ TAKING OUT "NO ALTA" ############################

## Drop rows where Discharge reason is "No Alta" or a transfer
df_two_binary = df_two_binary.drop(df_two_binary[df_two_binary["discharge_reason"] == "NO ALTA"].index)

#Change text (object) variables to be numbers
#Randomly done by LabelEncoder (we do this to because text data is more complicated to analyse)

df_two_binary['sex'] = LabelEncoder().fit_transform(df_two_binary['sex'])

df_two_binary['covid_diagnosis'] = LabelEncoder().fit_transform(df_two_binary['covid_diagnosis'])

df_two_binary['department'] = LabelEncoder().fit_transform(df_two_binary['department'])

df_two_binary['emergency_diagnosis']= LabelEncoder().fit_transform(df_two_binary['emergency_diagnosis'])

df_two_binary['emergency_dept'] = LabelEncoder().fit_transform(df_two_binary['emergency_dept'])

df_two_binary['discharge_reason'] = LabelEncoder().fit_transform(df_two_binary['discharge_reason'])


# DEFINING X AND Y

# Y = variables to be predicted (patient progression)
Y_binary = df_two_binary["discharge_reason"].values
# X = information we have about the patient
X_binary = df_two_binary.drop(labels = ["id", "discharge_reason"], axis=1)
features_list = list(X_binary.columns)

#Scale data to be between 0 and 1
scaler = MinMaxScaler()
scaler.fit(X_binary)
X_binary = scaler.transform(X_binary)

## PREDICTION WITH RF OVERSAMPLED BEST PARAMS ##
X_train_binary, X_test_binary, y_train_binary, y_test_binary = train_test_split(X_binary, Y_binary, test_size=0.2, random_state=32)

from imblearn.over_sampling import SMOTE
smoter = SMOTE(random_state=32)
X_over_binary, y_over_binary = smoter.fit_resample(X_train_binary, y_train_binary)
RF_best_params_binary = RandomForestClassifier(random_state=32)

RF_best_params_binary.fit(X_over_binary, y_over_binary)
RF_binary_prediction = RF_best_params_binary.predict(X_test_binary)

#Accuracy
print("Accuracy RF binary = ", metrics.accuracy_score(y_test_binary, RF_binary_prediction))
#0.82 (best)

# Let's check IoU
from keras.metrics import MeanIoU
num_classes = 2
IOU_RF = MeanIoU(num_classes=num_classes)  
IOU_RF.update_state(y_test_binary, RF_binary_prediction)
print("Mean IoU for KNN = ", IOU_RF.result().numpy())
#0.58 (best)

values = np.array(IOU_RF.get_weights()).reshape(num_classes, num_classes)
print(values)

class1_IoU = values[0,0]/(values[0,0] + values[0,1] + values[1,0])
class2_IoU = values[1,1]/(values[1,1] + values[1,0] + values[0,1])

print("IoU for class [Domicilio] is: ", class1_IoU)
print("IoU for class [Fallecimiento] is: ", class2_IoU)


# ROC CURVES
from sklearn.metrics import roc_curve, roc_auc_score

false_positive_rf, true_positive_rf, threshold1 = roc_curve(y_test, prediction_RF_over)
false_positive_svm, true_positive_svm, threshold2 = roc_curve(y_test, svm_prediction)
false_positive_knn, true_positive_knn, threshold3 = roc_curve(y_test, knn_prediction)

plt.subplots(1, figsize=(10,10))
plt.title('Receiver Operating Characteristic - DecisionTree')
plt.plot(false_positive_rf, true_positive_rf)
plt.plot([0, 1], ls="--")
plt.plot([0, 0], [1, 0] , c=".7"), plt.plot([1, 1] , c=".7")
plt.ylabel('True Positive Rate')
plt.xlabel('False Positive Rate')
plt.show()
  
  
  
  
  
  